# Autogenerated by nbdev

d = { 'settings': { 'branch': 'master',
                'doc_baseurl': '/theonerig/',
                'doc_host': 'https://Tom-TBT.github.io',
                'git_url': 'https://github.com/Tom-TBT/theonerig/tree/master/',
                'lib_path': 'theonerig'},
  'syms': { 'theonerig.core': {},
            'theonerig.database': {},
            'theonerig.eyetrack': {},
            'theonerig.leddome': {},
            'theonerig.modelling': {},
            'theonerig.plotting': {},
            'theonerig.processing': {},
            'theonerig.synchro.extracting': {},
            'theonerig.synchro.io': {},
            'theonerig.synchro.nested_stims': {},
            'theonerig.synchro.processing': {},
            'theonerig.testdata': {},
            'theonerig.utils': {}}}