# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"DataChunk": "00_core.ipynb",
         "ContiguousRecord": "00_core.ipynb",
         "RecordMaster": "00_core.ipynb",
         "Data_Pipe": "00_core.ipynb",
         "export_record": "00_core.ipynb",
         "import_record": "00_core.ipynb",
         "extend_sync_timepoints": "01_utils.ipynb",
         "align_sync_timepoints": "01_utils.ipynb",
         "resample_to_timepoints": "01_utils.ipynb",
         "link_sync_timepoints": "01_utils.ipynb",
         "flip_stimulus": "01_utils.ipynb",
         "flip_gratings": "01_utils.ipynb",
         "stim_to_dataChunk": "01_utils.ipynb",
         "phy_results_dict": "01_utils.ipynb",
         "spike_to_dataChunk": "01_utils.ipynb",
         "get_calcium_stack_lenghts": "01_utils.ipynb",
         "twoP_dataChunks": "01_utils.ipynb",
         "img_2d_fit": "01_utils.ipynb",
         "fill_nan": "01_utils.ipynb",
         "stim_inten_norm": "01_utils.ipynb",
         "group_direction_response": "01_utils.ipynb",
         "group_chirp_bumps": "01_utils.ipynb",
         "get_repeat_corrected": "01_utils.ipynb",
         "removeSlowDrift": "01_utils.ipynb",
         "time_shift_test_corr": "01_utils.ipynb",
         "cross_corr_with_lag": "01_utils.ipynb",
         "get_inception_generator": "01_utils.ipynb",
         "group_omitted_epochs": "01_utils.ipynb",
         "get_shank_channels": "01_utils.ipynb",
         "format_pval": "01_utils.ipynb",
         "stim_recap_df": "01_utils.ipynb",
         "eyetrack_stim_inten": "02_processing.ipynb",
         "saccade_distances": "02_processing.ipynb",
         "smooth_eye_position": "02_processing.ipynb",
         "process_sta_batch": "02_processing.ipynb",
         "staEst_fromBins": "02_processing.ipynb",
         "process_sta_batch_large": "02_processing.ipynb",
         "cross_correlation": "02_processing.ipynb",
         "corrcoef": "02_processing.ipynb",
         "flatten_corrcoef": "02_processing.ipynb",
         "stimulus_ensemble": "02_processing.ipynb",
         "process_nonlinearity": "02_processing.ipynb",
         "activity_histogram": "02_processing.ipynb",
         "cross_distances": "02_processing.ipynb",
         "cross_distances_sta": "02_processing.ipynb",
         "paired_distances": "02_processing.ipynb",
         "paired_distances_sta": "02_processing.ipynb",
         "direction_selectivity": "02_processing.ipynb",
         "wave_direction_selectivity": "02_processing.ipynb",
         "peri_saccadic_response": "02_processing.ipynb",
         "sigmoid": "03_modelling.ipynb",
         "gaussian": "03_modelling.ipynb",
         "sum_of_gaussian": "03_modelling.ipynb",
         "diff_of_gaussian": "03_modelling.ipynb",
         "gaussian_2D": "03_modelling.ipynb",
         "sum_of_2D_gaussian": "03_modelling.ipynb",
         "exponential_decay": "03_modelling.ipynb",
         "sin_exponent": "03_modelling.ipynb",
         "sinexp_sigm": "03_modelling.ipynb",
         "fit_sigmoid": "03_modelling.ipynb",
         "fit_spatial_sta": "03_modelling.ipynb",
         "fit_temporal_sta": "03_modelling.ipynb",
         "fit_chirp_am": "03_modelling.ipynb",
         "fit_chirp_freq_epoch": "03_modelling.ipynb",
         "fit_transiency": "03_modelling.ipynb",
         "repetition_quality_index": "03_modelling.ipynb",
         "onoff_transient_index": "03_modelling.ipynb",
         "DEFAULT_COLORS": "04_plotting.ipynb",
         "plot_2d_fit": "04_plotting.ipynb",
         "plot_tSTA_fit": "04_plotting.ipynb",
         "plot_chirpam_fit": "04_plotting.ipynb",
         "plot_chirp_freq_epoch_fit": "04_plotting.ipynb",
         "plot_transient_fit": "04_plotting.ipynb",
         "plot_nonlinearity_fit": "04_plotting.ipynb",
         "plot_ds_wheel": "04_plotting.ipynb",
         "plot_ds_wave_wheel": "04_plotting.ipynb",
         "plot_fl_bars": "04_plotting.ipynb",
         "plot_t_sta": "04_plotting.ipynb",
         "plot_chirp": "04_plotting.ipynb",
         "plot_spike_template": "04_plotting.ipynb",
         "plot_spike_template_kilo": "04_plotting.ipynb",
         "plot_spike_template_MEA": "04_plotting.ipynb",
         "plot_autocorrelogram": "04_plotting.ipynb",
         "plot_spike_amplitudes": "04_plotting.ipynb",
         "plot_spike_amplitudes_kilo": "04_plotting.ipynb",
         "plot_calcium_trace": "04_plotting.ipynb",
         "plot_stim_epochs_to_ephy": "04_plotting.ipynb",
         "plot_stim_epochs_to_calcium": "04_plotting.ipynb",
         "plot_cell_spatial": "04_plotting.ipynb",
         "plot_stim_recap_table": "04_plotting.ipynb",
         "plot_composed_A_masks": "04_plotting.ipynb",
         "plot_sta_positions": "04_plotting.ipynb",
         "plot_2d_sta": "04_plotting.ipynb",
         "plot_dome_flat": "04_plotting.ipynb",
         "plot_dome_checker": "04_plotting.ipynb",
         "plot_omitted_response": "04_plotting.ipynb",
         "plot_sta_pixelcorr": "04_plotting.ipynb",
         "plot_svd": "04_plotting.ipynb",
         "plot_nonlin": "04_plotting.ipynb",
         "plot_crosscorr_spikes_behav": "04_plotting.ipynb",
         "configure_pyplot_recap": "04_plotting.ipynb",
         "plot_recap_vivo_ephy": "04_plotting.ipynb",
         "plot_recap_vivo_ephy_kilo": "04_plotting.ipynb",
         "plot_recap_vivo_calcium": "04_plotting.ipynb",
         "plot_recap_vitro_ephy": "04_plotting.ipynb",
         "plot_recap_vivo_ephy_dome": "04_plotting.ipynb",
         "plot_recap_vivo_ephy_corr_behav": "04_plotting.ipynb",
         "plot_recap_wholeField": "04_plotting.ipynb",
         "plot_recap_wholeField_dome": "04_plotting.ipynb",
         "plot_recap_wholeField_vitroHiroki": "04_plotting.ipynb",
         "get_db_engine": "05_database.ipynb",
         "prompt_credentials": "05_database.ipynb",
         "get_record_essentials": "05_database.ipynb",
         "get_stim_params": "05_database.ipynb",
         "get_table": "05_database.ipynb",
         "stim_param_to_dict": "05_database.ipynb",
         "interpolate_screen_pos": "06_eyetrack.ipynb",
         "interpolate_checker_pos": "06_eyetrack.ipynb",
         "split_eye_events": "06_eyetrack.ipynb",
         "get_spherical_map": "06_eyetrack.ipynb",
         "apply_spherical_map": "06_eyetrack.ipynb",
         "atoi": "10_synchro.io.ipynb",
         "natural_keys": "10_synchro.io.ipynb",
         "filter_per_extension": "10_synchro.io.ipynb",
         "print_and_log": "10_synchro.io.ipynb",
         "print_info": "10_synchro.io.ipynb",
         "print_error": "10_synchro.io.ipynb",
         "get_offset": "10_synchro.io.ipynb",
         "logger": "10_synchro.io.ipynb",
         "DataFile": "10_synchro.io.ipynb",
         "read_header": "10_synchro.io.ipynb",
         "get_bytes_per_data_block": "10_synchro.io.ipynb",
         "read_qstring": "10_synchro.io.ipynb",
         "RHDFile": "10_synchro.io.ipynb",
         "H5File": "10_synchro.io.ipynb",
         "RawBinaryFile": "10_synchro.io.ipynb",
         "NumpyFile": "10_synchro.io.ipynb",
         "load_all_data": "10_synchro.io.ipynb",
         "load_all_data_adc": "10_synchro.io.ipynb",
         "load_all_data_dig_in": "10_synchro.io.ipynb",
         "load_all_data_both": "10_synchro.io.ipynb",
         "export_adc_raw": "10_synchro.io.ipynb",
         "export_dig_in_raw": "10_synchro.io.ipynb",
         "export_raw": "10_synchro.io.ipynb",
         "export_both_raw": "10_synchro.io.ipynb",
         "load_adc_raw": "10_synchro.io.ipynb",
         "load_digin_raw": "10_synchro.io.ipynb",
         "load_sync_raw": "10_synchro.io.ipynb",
         "get_QDSpy_logs": "11_synchro.extracting.ipynb",
         "QDSpy_log": "11_synchro.extracting.ipynb",
         "Stimulus": "11_synchro.extracting.ipynb",
         "get_synced_file": "14_synchro.nested_stims.ipynb",
         "unpack_stim_npy": "11_synchro.extracting.ipynb",
         "extract_spyking_circus_results": "11_synchro.extracting.ipynb",
         "extract_best_pupil": "11_synchro.extracting.ipynb",
         "stack_len_extraction": "11_synchro.extracting.ipynb",
         "get_thresholds": "12_synchro.processing.ipynb",
         "get_first_high": "12_synchro.processing.ipynb",
         "detect_frames": "12_synchro.processing.ipynb",
         "reverse_detection": "12_synchro.processing.ipynb",
         "extend_timepoints": "12_synchro.processing.ipynb",
         "error_check": "12_synchro.processing.ipynb",
         "cluster_frame_signals": "12_synchro.processing.ipynb",
         "cluster_by_epochs": "12_synchro.processing.ipynb",
         "cluster_by_list": "14_synchro.nested_stims.ipynb",
         "parse_time": "12_synchro.processing.ipynb",
         "get_position_estimate": "12_synchro.processing.ipynb",
         "match_starting_position": "12_synchro.processing.ipynb",
         "display_match": "12_synchro.processing.ipynb",
         "frame_error_correction": "12_synchro.processing.ipynb",
         "error_frame_matches": "12_synchro.processing.ipynb",
         "apply_shifts": "12_synchro.processing.ipynb",
         "shift_detection_conv": "12_synchro.processing.ipynb",
         "shift_detection_NW": "12_synchro.processing.ipynb",
         "chop_stim_edges": "12_synchro.processing.ipynb",
         "detect_calcium_frames": "12_synchro.processing.ipynb",
         "get_dome_positions": "13_leddome.ipynb",
         "as_cartesian": "13_leddome.ipynb",
         "as_spherical": "13_leddome.ipynb",
         "angular_distance": "13_leddome.ipynb",
         "build_wave_stimulus_array": "13_leddome.ipynb",
         "Quaternion": "13_leddome.ipynb",
         "get_waves_relative_position": "13_leddome.ipynb",
         "get_led_relative_position": "13_leddome.ipynb",
         "get_stim_ids": "14_synchro.nested_stims.ipynb",
         "get_synced_file_precompile": "14_synchro.nested_stims.ipynb",
         "get_stim_dict": "14_synchro.nested_stims.ipynb",
         "find_stim_onsets": "14_synchro.nested_stims.ipynb",
         "append_tread_movement": "14_synchro.nested_stims.ipynb",
         "append_to_reM": "14_synchro.nested_stims.ipynb",
         "master_loop": "14_synchro.nested_stims.ipynb",
         "load_vivo_2p": "99_testdata.ipynb"}

modules = ["core.py",
           "utils.py",
           "processing.py",
           "modelling.py",
           "plotting.py",
           "database.py",
           "eyetrack.py",
           "synchro/io.py",
           "synchro/extracting.py",
           "synchro/processing.py",
           "leddome.py",
           "synchro/nested_stims.py",
           "testdata.py"]

doc_url = "https://Tom-TBT.github.io/theonerig/"

git_url = "https://github.com/Tom-TBT/theonerig/tree/master/"

def custom_doc_links(name): return None
